import argparse
import os
import io
import sys
import json 
import numpy as np
import pandas as pd
import math
import cv2
import platform
import requests
import logging
from datetime import datetime
import shutil

# plot
from PIL import Image
from matplotlib import pyplot as plt

### coco mask to polygon ###
import base64
from pycocotools.coco import COCO
from skimage import measure
from io import BytesIO

### parallel download images ###
import uuid
from joblib import Parallel, delayed
from concurrent import futures

#### shapely ###
import shapely
import shapely.wkt
from shapely.ops import unary_union
from shapely.geometry import Polygon, MultiPoint
from shapely.geometry import Polygon

class Ecopia_Geocoding:
    '''
    Aquire building boundary lat long from giving address
    '''
    def __init__(self, address, Ecopia_Api_Key = 'cKr6EbB2YxXtNobPfBrB5'):
        self.address = address
        self.Ecopia_Api_Key = Ecopia_Api_Key

    def ecopia_addr_API(self,addr):
        '''
        Returns the API request result.
        @addr is a string of the full address.
        '''
        url = 'https://geocoding.ecopiatech.com/v4/geocoding_wkt?api_key=' + self.Ecopia_Api_Key + '&address=' + addr
        return requests.get(url, verify=True) # verify=False is necessary due to the security setting in Zscaler; when moving to Azure/production, switch to verify=True

    def ecopia_parcel_detail_API(self, parcel_id):
        '''
        Returns the API request result.
        @addr is a string of the parcel ID.
        '''
        url = 'https://geocoding.ecopiatech.com/v4/parcel_details?api_key=' + self.Ecopia_Api_Key + '&parcel_id=' + parcel_id
        return requests.get(url, verify=True) # verify=False is necessary due to the security setting in Zscaler; when moving to Azure/production, switch to verify=True

    def ecopia_best_full_ftpts(self, addr):
        '''
        Returns a tuple of a flag, a json object, and a list. 
        The flag is whether the geocoding is the most accurate level results.
        The json object is the results from the address inquiry.
        The list is a list of footprints, each of which is a dictionary with 3 keys: 'id', 'wkt', and 'location'. The first two values are strings. The location value is a dictionary with two keys, 'lat' and 'lon'.
        The first item in the list is from the geocoding app, where most likely the input address is, while the others are buildings (fully or partially) on the same parcel.
        The first item contains 'wkt', which is the 'orig_wkt' from the address API, and 'clipped_wkt', which is the wkt from the address API, so it's consistent with the rest of ftpts from the parcel API.
        @addr is a string of the full address.
        '''
        self.address = addr
        r1 = self.ecopia_addr_API(self.address)
        result1 = r1.json()

        ftpt1 = dict()
        ftpt1['id'] = result1['result']['id']
        ftpt1['wkt'] = result1['result']['orig_wkt']
        ftpt1['clipped_wkt'] = result1['result']['wkt']
        ftpt1['location'] = result1['result']['location']

        ftpts = [ftpt1]
        accuracy_flag = False
        r2 = ''
        confi_level = 0 

        # if ecopia can find the best footprint return it
        if result1['status'] == True and result1['result']["approximation"] == "":
            accuracy_flag = True
            confi_level = int(result1['result']['confi_level'])
            # Geocoding result is at the most accurate level
            if len(result1['result']['wkt'])>0:
                return (accuracy_flag, result1, ftpts, r2, confi_level)
            # Geocoding result is not accurate enough
            else:
                parcel_id = result1['result']['parcel_id']
                r2 = self.ecopia_parcel_detail_API(parcel_id).json()
                if 'footprints' in r2.keys(): # In some cases, the returned results have no 'footprints', e.g. when the parcle is too big and the result will show some error message
                    r2_result = r2['footprints'] 
                    if isinstance(r2_result, list): # in some cases, there is no footprints returned and the value can be None rather than a list (even an empty list)
                        ftpts.extend(r2_result)
                return (accuracy_flag, result1, ftpts, r2, confi_level)
        else: 
            return (accuracy_flag, result1, ftpts, r2, confi_level)
    
    def get_ftpts_box(self, ftpts):
        '''
        @ftpts is a list of dictionaries, in which key 'wkt' points to building ftpt polygons in (long, lat) coordinates.
        Returns a bounding box to get image that should contain all footprints from @ftpts in the format: (min_lat, min_long,  max_lat, max_long) 
        '''

        wkt_list = [ftpt['wkt'] for ftpt in ftpts if not isinstance(ftpt['wkt'], list)] # address API wkt value is a list, while parcel API wkt value is a string (only one polygon)
        wkt_list.extend([wkt for ftpt in ftpts for wkt in ftpt['wkt'] if isinstance(ftpt['wkt'], list)])

        polys = [shapely.wkt.loads(wkt) for wkt in wkt_list]
        bounds = [poly.bounds for poly in polys]
        longs = [b[0] for b in bounds]
        longs.extend([b[2] for b in bounds])
        lats = [b[1] for b in bounds]
        lats.extend([b[3] for b in bounds])

        if len(lats) < 2: return None # sometimes there are no wkts in Ecopia results

        min_lat = min(lats)
        max_lat = max(lats)
        min_long = min(longs)
        max_long = max(longs)
        # print([min_lat,min_long,  max_lat, max_long]) 
        return [min_lat, min_long, max_lat, max_long]

    
    def boundary_to_lat_lon_with_buffer(self, polyBounds, flag = 0):
        '''
        Returns a bounding box with a buffer to current polyBounds to get image
        polyBounds is in the format: (min_lat, min_long, max_lat, max_long) 
        output formate is format: (min_lat, min_long, max_lat, max_long) 
        flag = 0:
        buffer is a abslute value 0.00002 degrees
        flag = 1:
        buffer is either the 10% of the bounding box or 0.00002 degrees which is min
        '''
        if flag == 0:
            buffer_w = 0.00003 # 0.00002
            buffer_h = 0.00003 # 0.00002
        elif flag == 1: 
            buffer_percentage = 0.1
            abs_buffer = 0.00002
            # width and height
            width = abs(polyBounds[2]-polyBounds[0])
            height = abs(polyBounds[3]-polyBounds[1])

            buffer_w = min(buffer_percentage * width, abs_buffer)
            buffer_h = min(buffer_percentage * height, abs_buffer)

        # new corners lat and long
        top_left_lat = polyBounds[0] - buffer_h
        top_left_lon = polyBounds[1] - buffer_w
        bottom_right_lat = polyBounds[2] + buffer_h
        bottom_right_lon = polyBounds[3] + buffer_w

        # print(top_left_lat,top_left_lon,bottom_right_lat,bottom_right_lon)
        return [top_left_lat,top_left_lon,bottom_right_lat,bottom_right_lon]
    
    def address2boundary_lat_long_with_buffer(self):
        accuracy_flag, result1, ftpts, r2, confi_level = self.ecopia_best_full_ftpts(self.address)
        # print(ftpts)
        self.result1 = result1
        self.ftpts = ftpts
        self.r2 = r2
        if accuracy_flag:
            lat_lon_boundary_no_buffer = self.get_ftpts_box(ftpts)
            lat_lon_coord= self.boundary_to_lat_lon_with_buffer(lat_lon_boundary_no_buffer)
            return lat_lon_coord
        else:
            return []

    def address2boundary_lat_long_with_buffer_and_confi_level(self):
        accuracy_flag, result1, ftpts, r2, confi_level = self.ecopia_best_full_ftpts(self.address)
        # print(ftpts)
        self.result1 = result1
        self.ftpts = ftpts
        self.r2 = r2
        if confi_level > 6:
            lat_lon_boundary_no_buffer = self.get_ftpts_box(ftpts)
            lat_lon_coord= self.boundary_to_lat_lon_with_buffer(lat_lon_boundary_no_buffer)
            return lat_lon_coord, confi_level
        else:
            return []

############################################### Near Map
class Nearmap_Image:
    '''
    Aquire image giving boundary(min_long, min_lat, max_long, max_lat)
    and check date '%Y-%m-%d' '2022-05-01'
    '''
    def __init__(self, boundary, zoom_level = None, date=datetime.today().strftime('%Y-%m-%d'),
                 Nearmap_Api_Key = 'NGZkNzZmNTUtYWQ3Ny00NTdmLWIzMGEtZTIwZjk1OGZjNDM4'):
        self.boundary = boundary
        self.check_date = date
        self.zoom_level = zoom_level
        self.Nearmap_Api_Key = Nearmap_Api_Key
        # self.Today_date = datetime.today().strftime('%Y-%m-%d')
        # Today_date = '2022-05-01'
        # print(f'Today is: {self.date}.')

    def lat_lon_to_tile_coords(self, lat_deg, lon_deg, zoom):
        # From: https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames
        lat_rad = math.radians(lat_deg)
        n = 2.0 ** zoom
        xtile = int((lon_deg + 180.0) / 360.0 * n)
        ytile = int((1.0 - math.asinh(math.tan(lat_rad)) / math.pi) / 2.0 * n)
        return [xtile, ytile]

    def tile_coords_to_lat_lon(self, xtile, ytile, zoom, flag = 0):
        '''
        flag = 0 : NW-corner of the square
        flag = 1 : center
        flag = 2 : Other corner
        '''
        if flag == 1:
            xtile+=0.5
            ytile+=0.5

        if flag ==2:
            xtile+=1
            ytile+=1

        n = 2.0 ** zoom
        lon_deg = xtile / n * 360.0 - 180.0
        lat_rad = math.atan(math.sinh(math.pi * (1 - 2 * ytile / n)))
        lat_deg = math.degrees(lat_rad)
        return [lat_deg, lon_deg]
    
    def check_the_best_zoom_level_through_date(self):
        '''
        Return the best zoom level from the most closest date until the check date
        zoom_level: Uses the Google Maps Tile Coordinates.
        '''
        lat_deg =(self.boundary[0] + self.boundary[2])/2 #42.04508791988566
        lon_deg =(self.boundary[1] + self.boundary[3])/2 #-88.0559192506656
        
        url = 'https://api.nearmap.com/coverage/v2/point/'+str(lon_deg) +',' + str(lat_deg)+'?apikey='+ self.Nearmap_Api_Key +'&limit=1' + '&until=' + self.check_date
        # print(url)
        zoom_request = requests.get(url, verify=True)
        nearmap_meta = zoom_request.json()
        self.image_date = nearmap_meta['surveys'][0]['captureDate']
        # print(nearmap_meta['surveys'])
        
        if self.zoom_level is None:
            # obtain the best scale
            if isinstance(nearmap_meta['surveys'][0]['resources']['tiles'][0]['scale'], int):
                for tile_meta in nearmap_meta['surveys'][0]['resources']['tiles']:
                    if tile_meta['type']=='Vert':
                        zoom_level = tile_meta['scale']
                        break
                    else:
                        zoom_level = tile_meta['scale']
            else:
                zoom_level = 21
            self.zoom_level = zoom_level
        return self.zoom_level, self.image_date

    def parallelize_download_tiles(self, i, j):
        url = 'https://api.nearmap.com/tiles/v3/Vert/'+ str(self.zoom_level)+ '/' + str(i) + '/' + str(j) + '.jpg?apikey=' + self.Nearmap_Api_Key + '&until=' + self.image_date
        # print(url)
        r = requests.get(url, verify=True) # verify=False is necessary due to the security setting in Zscaler; when moving to Azure/production, switch to verify=True
        # bytes to image
        plt_image = Image.open(BytesIO(r.content))

        # transfer to opencv format
        opencvImage = cv2.cvtColor(np.array(plt_image), cv2.COLOR_RGB2BGR)
        return i,j,opencvImage
    
    def multithread_download_tiles(self, i, j):
        url = 'https://api.nearmap.com/tiles/v3/Vert/'+ str(self.zoom_level)+ '/' + str(i) + '/' + str(j) + '.jpg?apikey=' + self.Nearmap_Api_Key + '&until=' + self.image_date
        # print(url)
        r = requests.get(url, verify=True) # verify=False is necessary due to the security setting in Zscaler; when moving to Azure/production, switch to verify=True
        # bytes to image
        plt_image = Image.open(BytesIO(r.content))

        # transfer to opencv format
        self.new_image_multithread[(j-self.y0)*256: (j-self.y0)*256+256, (i-self.x0)*256: (i-self.x0)*256+256, :] = cv2.cvtColor(np.array(plt_image), cv2.COLOR_RGB2BGR)

    
    def lat_lon_to_opencv_image(self):
        '''
        lat_long_coor is boundary(min_long, min_lat, max_long, max_lat)
        
        Return a reconstructe Nearmap tile images. Each tile image is 256*256.
        Based on the lat_long, we compute the number of tiles need to be used.
        
        tile_resource_type:
            The resource type for the requested tiles. The available values are:
            Vert - for vertical imagery
            North - for North panorama imagery
            South - for South panorama imagery
            East - for East panorama imagery
            West - for West panorama imagery
            Note: the tileResourceType values are case sensitive.
        '''
        format = "jpg"
        tile_resource_type = "Vert"
 
        # Convert lat, lon and zoom to x,y,z
        x0, y1 = self.lat_lon_to_tile_coords(lat_deg = self.boundary[0], lon_deg=self.boundary[1], zoom= self.zoom_level)
        x1, y0 = self.lat_lon_to_tile_coords(lat_deg = self.boundary[2], lon_deg=self.boundary[3], zoom= self.zoom_level)
        self.x0 = x0
        self.y1 = y1
        self.x1 = x1
        self.y0 = y0
        logging.info(f"x:{x0},y:{y1}")
        logging.info(f"x:{x1},y:{y0}")
        self.nw_lat_long = self.tile_coords_to_lat_lon(x0,y0,self.zoom_level, flag=0)
        self.se_lat_long = self.tile_coords_to_lat_lon(x1,y1,self.zoom_level, flag=2)  

        self.new_image_multithread = np.zeros((256*(y1-y0+1),256*(x1-x0+1),3), np.uint8)
        with futures.ThreadPoolExecutor(max_workers=100) as executor:
            task_list = [executor.submit(self.multithread_download_tiles, i, j) for i in range(x0,x1+1) for j in range(y0,y1+1)]

        return self.new_image_multithread
    
    def boundary2image(self):
        zoom, date = self.check_the_best_zoom_level_through_date()
        full_image = self.lat_lon_to_opencv_image()
        return full_image, zoom, date

### Functions for ground sample distance computation  ###  
def pixels_to_meters(latitude=40, zoom_level=21):
    equatorial_circumference = 2 * math.pi * 6378137
    meter_per_pixels = equatorial_circumference * math.cos(math.radians(latitude)) / math.pow(2,zoom_level+8)
    return meter_per_pixels

def image_download(input_address, name, file_root_path):
    excel_file_path = os.path.join(file_root_path,'roof_score_new_20230101.csv')
    image_folder_name = os.path.splitext(os.path.basename(excel_file_path))[0]
    demoFolder = image_folder_name
    # make a image folder by excel name
    if not(os.path.isdir(image_folder_name)):
        os.makedirs(image_folder_name)
    else:
        shutil.rmtree(image_folder_name)
        os.makedirs(image_folder_name)
        

    # load the excel file
    excel_file = pd.read_csv(excel_file_path)
    # updating address value
    excel_file.loc[0, 'location_Address'] = input_address
    excel_file.loc[0, 'InsuredName'] = name
    excel_file.to_csv(excel_file_path, index=False)

    full_address = excel_file["location_Address"].tolist()
    insured_name = excel_file["InsuredName"].tolist()
    policy_expiration_date = excel_file["PolicyExpirationDate"].tolist()
    Today_date = datetime.today().strftime('%Y-%m-%d')
    
    # go through each address name
    image_name_list = []
    csv_image_boundary = []
    csv_image_info = []
    csv_image_level = []
    csv_image_date = []
    csv_gsd = []
    # output_address = []
    for i in range(len(full_address)):
        image_text = ''
        address = full_address[i]
        # geocoding
        Ecopia_API = Ecopia_Geocoding(address)
        try:
            address_boundary = Ecopia_API.address2boundary_lat_long_with_buffer()
        except:
            # image_name = address + '_'+ Today_date
            image_name = address
            logging.info(f'index:{i}, address:{address}')
            image_text += 'The Ecopia Geocoding is not avaliable.'
            logging.info(image_text)
            image_name_list.append(image_name)
            csv_image_boundary.append('')
            csv_image_info.append(image_text)
            csv_image_level.append('')
            csv_image_date.append('')
            csv_gsd.append('')
            continue
        if address_boundary == []:
            # image_name = address + '_'+ Today_date
            image_name = address
            logging.info(f'index:{i}, address:{address}')
            image_text += 'Ecopia Geocoding is not accurate.'
            logging.info(image_text)
            image_name_list.append(image_name)
            csv_image_boundary.append('')
            csv_image_info.append(image_text)
            csv_image_level.append('')
            csv_image_date.append('')
            csv_gsd.append('')
            continue

        # image acquire
        check_date = Today_date

        # zoom level 21 is better than others because the training set almost all level21
        Nearmap_API = Nearmap_Image(address_boundary,zoom_level=21, date=check_date)
        try:
            # 0	1.0	111 km
            # 1	0.1	11.1 km
            # 2	0.01	1.11 km
            # 3	0.001	111 m
            # 4	0.0001	11.1 m
            # 5	0.00001	1.11 m
            # 6	0.000001	0.111 m
            # 7	0.0000001	1.11 cm
            # 8	0.00000001	1.11 mm
            if address_boundary[2] - address_boundary[0] > 0.003 or address_boundary[3] - address_boundary[1] > 0.003:
                # image_name = address + '_'+ Today_date
                image_name = address
                logging.info(f'index:{i}, address:{address}')
                image_text += f'image is too big {address_boundary}. Human check!'
                logging.info(image_text)
                image_name_list.append(image_name)
                csv_image_boundary.append(address_boundary)
                csv_image_info.append(image_text)
                csv_image_level.append('')
                csv_image_date.append('')
                csv_gsd.append('')
                continue

            image, zoom_level, date = Nearmap_API.boundary2image()
            # image_name = address + '_' + str(zoom_level)+ '_'+ date
            image_name = address
            logging.info(f'index:{i}, address:{address}')
            image_text = image_text + f'[{Nearmap_API.x0}, {Nearmap_API.y1}, {Nearmap_API.x1}, {Nearmap_API.y0}]'
            logging.info(image_text)
            cv2.imwrite(os.path.join(image_folder_name, image_name) + '.jpg', image)
            image_name_list.append(image_name)
            csv_image_boundary.append(address_boundary)
            csv_image_info.append(image_text)
            csv_image_level.append(int(zoom_level))
            csv_image_date.append(date)
            gsd = pixels_to_meters(latitude = (address_boundary[2] + address_boundary[0])/2, zoom_level = zoom_level)
            csv_gsd.append(gsd)
            continue
            
        # # image = np.zeros((100,100,3), np.uint8) 
        except:
            logging.info(f'index:{i}, address:{address}')
            # image_name = address + '_'+ Today_date
            image_name = address
            image_text += 'Nearmap image is not available here.'
            logging.info(image_text)
            image_name_list.append(image_name)
            csv_image_boundary.append(address_boundary)
            csv_image_info.append(image_text)
            csv_image_level.append('')
            csv_image_date.append('') 
            csv_gsd.append('')
    # logging.info(image_name_list)
    # logging.info(insured_name)
    # logging.info(policy_expiration_date)
    # logging.info(csv_image_date)
    # logging.info(csv_image_level)
    # logging.info(csv_image_boundary)
    # logging.info(csv_image_info)
    # logging.info(csv_gsd)
    out_csv = pd.DataFrame({'Address':image_name_list,
                            'InsuredName':insured_name,
                            'PolicyExpirationDate':policy_expiration_date,
                            'Image date': csv_image_date,
                            'Image level': csv_image_level,
                            'Image boundary': csv_image_boundary,
                            'Image info': csv_image_info,
                            'Image gsd (meters/pixels)': csv_gsd})
    out_csv.to_csv(os.path.join(image_folder_name,'image.csv'), index=False, header=True) 



### report ###
import reportlab
from reportlab.pdfgen import canvas
from reportlab.pdfbase.ttfonts import TTFont
from reportlab.pdfbase import pdfmetrics
from reportlab.lib import colors

from yolo_roof.detect_functions import *
#### shapely ###
import shapely
import shapely.wkt
from shapely.ops import unary_union
from shapely.geometry import Polygon, MultiPoint
from shapely.geometry import Polygon

## PDF
import reportlab
from reportlab.pdfgen import canvas
from reportlab.pdfbase.ttfonts import TTFont
from reportlab.pdfbase import pdfmetrics
from reportlab.lib import colors

## ZIP
import zipfile

# email
from email.mime.text import MIMEText
from email.header import Header
from smtplib import SMTP_SSL
from email.mime.multipart import MIMEMultipart

### Function for target address center lat long computation ###
def image_info_to_center_lat_lon(image_info, zoom, flag = 1):
    '''
    image_info is '[topleft tile x,
                    topleft tile y,
                    bottom right tile x,
                    bottom right tile y]'
    flag = 0 : NW-corner of the square
    flag = 1 : center
    flag = 2 : Other corner
    '''
    image_info_array = image_info.split(',')
    if flag == 0:
        xtile = (int(image_info_array[0][1:]) + int(image_info_array[2]))//2
        ytile = (int(image_info_array[1]) + int(image_info_array[3][:-1]))//2
    
    elif flag == 1:
        xtile = (int(image_info_array[0][1:]) + int(image_info_array[2]) + 1)//2
        ytile = (int(image_info_array[1]) + int(image_info_array[3][:-1]) + 1)//2

    elif flag ==2:
        xtile = (int(image_info_array[0][1:]) + int(image_info_array[2]) + 2)//2
        ytile = (int(image_info_array[1]) + int(image_info_array[3][:-1]) + 2)//2
    else:
        print('flag is not defined. Using NW-corner of the square')
        xtile = (int(image_info_array[0][1:]) + int(image_info_array[2]))//2
        ytile = (int(image_info_array[1]) + int(image_info_array[3][:-1]))//2

    n = 2.0 ** zoom
    lon_deg = xtile / n * 360.0 - 180.0
    lat_rad = math.atan(math.sinh(math.pi * (1 - 2 * ytile / n)))
    lat_deg = math.degrees(lat_rad)
    return [lat_deg, lon_deg]

def merge_polygones(image_folder,json_file,trainingDataset):
    allColors = [(255,0,0), (0,255,0), (0,0,255), (255,255,0), (255,0,255), (0,255,255), (75,125,255), (125,75,255), (75,255,125), (125,255,75), (255,125,75), (255,75,125)]*8
    filename = str(uuid.uuid4())
    register_coco_instances(filename, {'thing_colors':allColors}, json_file, image_folder)
    metadata = MetadataCatalog.get(filename)
    dicts = DatasetCatalog.get(filename)
    model_folder = os.path.join(image_folder,'prediction_' + trainingDataset)
    print(f"Merging {trainingDataset} polygons")
    for dic in dicts:
        img = utils.read_image(dic["file_name"], "RGB")
        visualizer = Visualizer(img, metadata=metadata, instance_mode=ColorMode.SEGMENTATION)
        vis = visualizer.draw_dataset_dict(dic)
        out_filename = os.path.join(os.path.splitext(os.path.basename(dic["file_name"]))[0],trainingDataset +'_merged.jpg')
        
        out_filename2_extention = os.path.splitext(os.path.basename(dic["file_name"]))[1]
        out_filename2 = os.path.basename(dic["file_name"]).replace(out_filename2_extention,'_merged'+out_filename2_extention)
        vis.save(os.path.join(image_folder,out_filename))
        # vis.save(os.path.join(model_folder,out_filename2)) # model folder is not necessary
    print("Merging Finished.")
    return vis

def generate_referral_json(demoFolder, coco_json, new_cat, new_json_file_name):
    catIds = coco_json.getCatIds(catNms=[], supNms=[], catIds=[])
    # Obtain the category names
    categories = coco_json.loadCats(catIds)
    new_categories = [x['name'] for x in categories if x['name'] in new_cat]
    new_catIds = coco_json.getCatIds(catNms=new_categories, supNms=[], catIds=[])
    print(new_catIds)
    imgIds = coco_json.getImgIds()
    
    new_json_file = os.path.join(demoFolder, new_json_file_name)
    
    new_images = []
    new_annotations = []
    ann_id = 1
    new_image_id = 1

    for img_id in range(len(imgIds)):
        # get image information
        coco_img_info = coco_json.loadImgs(imgIds[img_id])[0]
        # get the raw image annotations Id
        annIds = coco_json.getAnnIds(imgIds=coco_img_info['id'])
        # get annotations from annotation Id
        annotations = coco_json.loadAnns(annIds)

        ann_start = ann_id
        for ann in annotations:
            # print(ann)
            # class selection
            if ann['category_id'] in new_catIds:
                
                # append annotation to coco jason file list
                new_annotations.append({"id": ann_id,
                                        "image_id": new_image_id,
                                        "category_id": ann['category_id'],
                                        "iscrowd": 0,
                                        # "area": ann['area'],
                                        # "bbox": ann['bbox'],
                                        "segmentation": ann['segmentation']
                                        })
                # get new annotation id
                ann_id += 1

        # get new image id
        new_image_id += 1
        # append image information to coco jason file list
        new_images.append(coco_img_info)

    info = {"year": 2022,
            "version": "1.0",
            "description": "Phase 2 data and coco json file",
            "contributor": "Ruixu Liu, Delin Shen, Aaron Lee and Qiang Wang",
            }

    json_data = {"info": info,
                 "categories": categories,
                 "images": new_images,
                 "annotations": new_annotations
                 }

    with open(new_json_file, "w") as jsonfile:
        json.dump(json_data, jsonfile, sort_keys=True, indent=4)


        
def generate_referral_damage_json(demoFolder, coco_json, roof_boundary_json, new_cat, new_json_file_name):
    catIds = coco_json.getCatIds(catNms=[], supNms=[], catIds=[])
    # Obtain the category names
    categories = coco_json.loadCats(catIds)
    new_categories = [x['name'] for x in categories if x['name'] in new_cat]
    new_catIds = coco_json.getCatIds(catNms=new_categories, supNms=[], catIds=[])
    print(new_catIds)
    imgIds = coco_json.getImgIds()
    
    new_json_file = os.path.join(demoFolder, new_json_file_name)
    
    new_images = []
    new_annotations = []
    ann_id = 1
    new_image_id = 1

    for img_id in range(len(imgIds)):
        # get image information
        coco_img_info = coco_json.loadImgs(imgIds[img_id])[0]
        # get the raw image annotations Id
        annIds = coco_json.getAnnIds(imgIds=coco_img_info['id'])
        # get annotations from annotation Id
        annotations = coco_json.loadAnns(annIds)
        
        roof_poly = Polygon()
        # get image information
        coco_img_info = roof_boundary_json.loadImgs(imgIds[img_id])[0]
        # get the raw image annotations Id
        annIds = roof_boundary_json.getAnnIds(imgIds=coco_img_info['id'])
        # get annotations from annotation Id
        roof_annotations = roof_boundary_json.loadAnns(annIds)
        for ann in roof_annotations:
            for seg in ann['segmentation']:
                mask = np.array(seg).reshape(len(seg) // 2, 2)
                poly = Polygon(mask)
                roof_poly = roof_poly.union(poly)
        
        # roof_poly = roof_poly.buffer(-5.0)
        
        ann_start = ann_id
        for ann in annotations:
            # print(ann)
            # class selection
            if ann['category_id'] in new_catIds:
                current_poly = Polygon()
                for seg in ann['segmentation']:
                    mask = np.array(seg).reshape(len(seg) // 2, 2)
                    new_poly = Polygon(mask)
                    current_poly = current_poly.union(new_poly)

                if not current_poly.intersects(roof_poly):
                    continue
                else:
                    intersection_poly = current_poly.intersection(roof_poly)
                    
                if intersection_poly.geom_type == 'MultiPolygon':
                    for current_mask in intersection_poly:
                        if current_mask.area < 25:
                            continue
                        current_polygon_list = current_mask.exterior.coords
                        coco_segmentation_format = np.array(current_polygon_list).reshape(1, len(current_mask.exterior.coords) * 2).tolist()
                        # append annotation to coco jason file list
                        new_annotations.append({"id": ann_id,
                                                "image_id": new_image_id,
                                                "category_id": ann['category_id'],
                                                "iscrowd": 0,
                                                # "area": ann['area'],
                                                # "bbox": ann['bbox'],
                                                "segmentation": coco_segmentation_format
                                                })
                        # get new annotation id
                        ann_id += 1
                elif intersection_poly.geom_type == 'Polygon':
                    if intersection_poly.area < 100:
                        continue
                    current_polygon_list = intersection_poly.exterior.coords
                    coco_segmentation_format = np.array(current_polygon_list).reshape(1, len(current_polygon_list) * 2).tolist()
                    # append annotation to coco jason file list
                    new_annotations.append({"id": ann_id,
                                            "image_id": new_image_id,
                                            "category_id": ann['category_id'],
                                            "iscrowd": 0,
                                            # "area": ann['area'],
                                            # "bbox": ann['bbox'],
                                            "segmentation": coco_segmentation_format
                                            })
                    # get new annotation id
                    ann_id += 1             

        # get new image id
        new_image_id += 1
        # append image information to coco jason file list
        new_images.append(coco_img_info)

    info = {"year": 2022,
            "version": "1.0",
            "description": "Phase 2 data and coco json file",
            "contributor": "Ruixu Liu, Delin Shen, Aaron Lee and Qiang Wang",
            }

    json_data = {"info": info,
                 "categories": categories,
                 "images": new_images,
                 "annotations": new_annotations
                 }

    with open(new_json_file, "w") as jsonfile:
        json.dump(json_data, jsonfile, sort_keys=True, indent=4)


class ReferralPoints:
    '''
    Read in referral rules and generate the csv format:
    Address    Total points    each referral elemten
      aaa          3              3  
    '''

    def __init__(self, referral_file, index):
        with open(referral_file) as f1:
            self.ref_json = json.load(f1)
        self.score = {'Address': index}
        self.score.update({'Total points': 0})
        self.score.update(dict((score_cat, 0) for score_cat in self.ref_json))
        self.score.update({'Roof area (pixels)': 0})

        self.score_cat = [x for x in self.ref_json]

        self.buffer = dict((score_cat,{'area_percentage': 0, 'count': 0}) for score_cat in self.ref_json)
        self.buffer.update({'Roof_boundary': 0})
        
        # print(self.score_cat)
        # print(self.buffer)

    def set_points(self, cat, points=0):
        self.score[cat] = points

    def update_total_points(self):
        for cat in self.score_cat:
            score = self.score[cat]
            self.score['Total points'] += score
            # print(f'{cat}:, {score}')

    def obtation_roof_boundary(self, predict_result,image_id):
        # get the image annotations ids
        annIds = predict_result.getAnnIds(imgIds=image_id)
        # get annotations from annotation ids
        annotations = predict_result.loadAnns(annIds)
        self.roof_poly = Polygon()
        for ann in annotations:
            for seg in ann['segmentation']:
                mask = np.array(seg).reshape(len(seg) // 2, 2)
                poly = Polygon(mask)
                self.roof_poly = self.roof_poly.union(poly)
                poly_area = poly.area
                self.buffer['Roof_boundary'] += poly_area
    
        ### compute roof area
        self.score['Roof area (pixels)'] = poly_area
        self.roof_poly = self.roof_poly.buffer(5.0) #add a buffer for HVAC which are located on the ground
        # aaa = self.buffer['Roof_boundary']
        # # print(poly)
        # print(f'roof {aaa*0.0002}')
        
    def update_buffer_value(self, predict_result, labels, image_id):
        # get the image annotations ids
        annIds = predict_result.getAnnIds(imgIds=image_id)
        # get annotations from annotation ids
        annotations = predict_result.loadAnns(annIds)
        # print(len(annotations))

        # update referral buffer
        for ann in annotations:
            if labels[ann['category_id']] in self.score_cat:
                                
                # if this instance is detected 'area_percentage' + current mask area
                for seg in ann['segmentation']:
                    mask = np.array(seg).reshape(len(seg) // 2, 2)
                    poly = Polygon(mask)
                    self.current_poly = poly
                    if not self.current_poly.intersects(self.roof_poly):
                        continue
                    poly_area = poly.area
                    poly_percentage = poly_area / (self.buffer['Roof_boundary'] + 1)
                    self.buffer[labels[ann['category_id']]]['area_percentage'] += poly_percentage
                    
                    # print(poly)
                    # class_label = labels[ann['category_id']]
                    # print(f'{class_label} poly_area {poly_area}')

                if not self.current_poly.intersects(self.roof_poly):
                    continue
                # if this instance is detected 'count' + 1
                self.buffer[labels[ann['category_id']]]['count'] += 1
                    
        # for each cat
        for referral_cat in self.score_cat:
            # for each value section
            for referral_value in self.ref_json[referral_cat]:
                # for each referral type
                for refferal_type in self.buffer[referral_cat]:
                    # print(refferal_type)
                    if refferal_type == referral_value['type']:
                        # if current value is not 0 and current value is in the value section
                        # and the new value is larger than the old value
                        # update the value
                        if self.buffer[referral_cat][refferal_type]!=0:
                            if referral_value['min'] <= self.buffer[referral_cat][refferal_type] < referral_value['max']:
                                if referral_value['point'] > self.score[referral_cat]:
                                    self.set_points(referral_cat, points=referral_value['point'])
                # print(referral_cat)
                # print(self.buffer[referral_cat])  

def generate_report(file_root_path):
    demoFolder = 'roof_score_new_20230101'
    pred_damage_file = os.path.join(demoFolder,'prediction_Roof_damage/Roof_damage_all_in_one.json')
    pred_equip_file = os.path.join(demoFolder, 'prediction_Roof_equipment/Roof_equipment_all_in_one.json')
    pred_boundary_file =os.path.join(demoFolder, 'prediction_Roof_boundary/Roof_boundary_all_in_one.json')

    # read in referral rules and create corresponding functions
    referral_label_file = os.path.join(file_root_path,'referral_labels_20221006.json')
    with open(referral_label_file) as f1:
        ref_json = json.load(f1)
    cat = [x for x in ref_json]
    # Load json file
    pred_damage = COCO(pred_damage_file)
    pred_equip = COCO(pred_equip_file)
    pred_boundary = COCO(pred_boundary_file)
    generate_referral_damage_json(demoFolder, pred_damage,pred_boundary,cat,'Referral_damage.json')
    generate_referral_json(demoFolder, pred_equip,cat,'Referral_equipment.json')
    vis_out = merge_polygones(demoFolder,os.path.join(demoFolder, 'Referral_damage.json'),'Damage')
    vis_out = merge_polygones(demoFolder,os.path.join(demoFolder, 'Referral_equipment.json'),'Equipment')

    ################################### Referral
    referral_output = []
    # Obtain the category ID numbers
    catIds = pred_damage.getCatIds()
    # Obtain the category names
    categories = pred_damage.loadCats(catIds)
    # Id2catname
    Damage_labels = [value['name'] for value in categories]

    # Obtain the category ID numbers
    catIds = pred_equip.getCatIds()
    # Obtain the category names
    categories = pred_equip.loadCats(catIds)
    # Id2catname
    Equipment_labels = [value['name'] for value in categories]

    for damage_img_id, equp_image_id in zip(pred_damage.getImgIds(), pred_equip.getImgIds()):
        assert pred_damage.loadImgs(damage_img_id) == pred_damage.loadImgs(equp_image_id)
        # get image information
        coco_img_info = pred_damage.loadImgs(damage_img_id)[0]
        address = os.path.splitext(coco_img_info['file_name'])[0]
        print(address)
        # zoom_level = int(address[-13:-11])

        # using :-14 to only keep address no zoom level and date
        # new_referral = ReferralPoints(referral_label_file, index=address[:-14])
        new_referral = ReferralPoints(referral_label_file, index=address) 

        # Roof Boundary
        # new_referral.buffer['Roof boundary'] = np.sqrt(coco_img_info['height'] * coco_img_info['width'])
        # new_referral.buffer['Roof_boundary'] = 0.5* coco_img_info['height'] * coco_img_info['width']
        # print(coco_img_info['height'] * coco_img_info['width'])
        new_referral.obtation_roof_boundary(pred_boundary, image_id = coco_img_info['id'])


        # Damage
        #################################################
        new_referral.update_buffer_value(pred_damage, Damage_labels, image_id = coco_img_info['id'] )
        # if zoom_level==22:
        #     print(new_referral.buffer['Roof_boundary']*0.0002)
        #     print(new_referral.buffer['Water pooling']['area_percentage']*new_referral.buffer['Roof_boundary']*0.0002)
        # else:
        #     print(new_referral.buffer['Roof_boundary']*0.00085)
        #     print(new_referral.buffer['Water pooling']['area_percentage']*new_referral.buffer['Roof_boundary']*0.00085)

        # Equipment
        #################################################
        new_referral.update_buffer_value(pred_equip, Equipment_labels, image_id = coco_img_info['id'])

        new_referral.update_total_points()
        referral_output.append(new_referral.score)
        df = pd.DataFrame.from_dict([new_referral.score])
        df.to_csv(os.path.join(demoFolder,address+'/referral.csv'), index=False, header=True)

    df = pd.DataFrame.from_dict(referral_output)
    df.to_csv(os.path.join(demoFolder,'all_referral.csv'), index=False, header=True)
    image_csv = pd.read_csv(os.path.join(demoFolder,"image.csv"))
    referral_csv = pd.read_csv(os.path.join(demoFolder,"all_referral.csv"))

    image_csv.merge(referral_csv,how='outer',on='Address').to_csv(os.path.join(demoFolder, 'all_address.csv'), index=False, header=True)

    ################################################################# PDF
    image_csv = pd.read_csv(os.path.join(demoFolder,'image.csv'))
    referral_csv = pd.read_csv(os.path.join(demoFolder,'all_referral.csv'))
    all_address_csv = pd.read_csv(os.path.join(demoFolder, 'all_address.csv'))
    # Today_date = datetime.today().strftime('%Y-%m-%d')
    # # print(Today_date)
    full_address = all_address_csv["Address"].tolist()
    insured_name = all_address_csv["InsuredName"].tolist()
    image_date = all_address_csv["Image date"].tolist()
    image_info = all_address_csv["Image info"].tolist()
    zoom_level = all_address_csv["Image level"].tolist()

    key_name = all_address_csv.columns
    key_name_index = all_address_csv.columns.get_loc('Total points')
    gsd_key = all_address_csv.columns.get_loc('Image gsd (meters/pixels)')
    pixels_area_key = all_address_csv.columns.get_loc('Roof area (pixels)')
    # print(key_name[key_name_index])

    for address_index in range(len(full_address)):
        address = full_address[address_index]
        row = all_address_csv.iloc[address_index]
        print(f'index:{address_index}, address:{address}')
        if np.isnan(row[key_name_index]):
            continue
        # initializing variables with values
        url_zoom_level = zoom_level[address_index]
        cat_lon_coord = image_info_to_center_lat_lon(image_info[address_index], url_zoom_level)
        url_lat = cat_lon_coord[0]
        url_lon = cat_lon_coord[1]
        url_date = image_date[address_index].replace('-','')
        url = f'https://apps.nearmap.com/maps/#/@{url_lat},{url_lon},{url_zoom_level}z,0d/V/{url_date}'

        fileName = os.path.join(demoFolder,address+'.pdf')
        documentTitle = f'Refferal'
        title = f'{insured_name[address_index]}'
        subTitle = address

        Damage_image_name = os.path.join(demoFolder,address+'/Damage_merged.jpg')
        Equipment_image_name = os.path.join(demoFolder,address+'/Equipment_merged.jpg')

        pdf = canvas.Canvas(fileName)
        pdf.setTitle(documentTitle)
        pdf.setFont("Times-Roman", 14)
        pdf.drawCentredString(300, 800, title)

        pdf.setFillColorRGB(0, 0, 0)
        pdf.setFont("Times-Roman", 12)
        pdf.drawCentredString(290, 770, subTitle)

        # drawing a line
        pdf.line(30, 760, 550, 760)

        # information
        text = pdf.beginText(40, 740)
        text.setFont("Courier", 12)
        text.setFillColor(colors.black)
        text.textLine(f'Image date: {image_date[address_index]}')

        ######################################### 100 points
        # Range	    3	   3	        3	          3	                 3	           1	       3
        # weight	1	   2.5	        2.5	          1	                 1	           1	       1
        # Other damage	Staining	Water pooling	Skylight	HVAC/Cooling tower	Pipeline	Solar panel


        total_point = 0
        if row[key_name_index]!=0:
            for i in range(key_name_index+1, len(key_name)):
                if key_name[i] == 'Other damage':
                    total_point += row[i]*1
                elif key_name[i] == 'Staining':
                    total_point += row[i]*2.5
                elif key_name[i] == 'Water pooling':
                    total_point += row[i]*2.5
                elif key_name[i] == 'Skylight':
                    total_point += row[i]*1
                elif key_name[i] == 'HVAC/Cooling tower':
                    total_point += row[i]*1
                elif key_name[i] == 'Pipeline':
                    total_point += row[i]*1
                elif key_name[i] == 'Solar panel':
                    total_point += row[i]*1
            total_point = 100 - total_point * 5
            if total_point> 0 :
                line= f'Total scores: {total_point}'
                text.textLine(line)
                text.setFillColor(colors.blue)
            else:
                line = 'Total scores: 0'
                text.textLine(line)
                text.setFillColor(colors.blue)
        else:
            line = 'Total scores: 100'
            text.textLine(line)
            text.setFillColor(colors.blue)

        for i in range(key_name_index+1, len(key_name)):
            if i == pixels_area_key:
                gsd_area = row[gsd_key] * row[gsd_key] * row[pixels_area_key]
                text.textLine(f'Roof area: {int(gsd_area)} m\u00b2')
                continue
            if int(row[i]) == 0:
                line = key_name[i] + ': ' + 'Not detected'
                text.textLine(line)
            elif int(row[i]) == 1:
                if key_name[i] == 'Pipline' or 'Solar panel': 
                    line = key_name[i] + ': ' + 'Detected'
                    text.textLine(line)
                    continue
                if key_name[i] == 'Staining' or 'Water pooling' or 'Other damage':
                    line = key_name[i] + ': ' + 'Minor'
                    text.textLine(line)
                    continue
                if key_name[i] == 'Skylight' or 'HVAC/Cooling tower':
                    line = key_name[i] + ': ' + 'Minimal'
                    text.textLine(line)
                    continue
            elif int(row[i]) == 2:
                if key_name[i] == 'Pipline' or 'Solar panel': 
                    line = key_name[i] + ': ' + 'Detected'
                    text.textLine(line)
                    continue
                if key_name[i] == 'Staining' or 'Water pooling' or 'Other damage':
                    line = key_name[i] + ': ' + 'Moderate'
                    text.textLine(line)
                    continue
                if key_name[i] == 'Skylight' or 'HVAC/Cooling tower':
                    line = key_name[i] + ': ' + 'A few'
                    text.textLine(line)
                    continue
            else:
                if key_name[i] == 'Pipline' or 'Solar panel': 
                    line = key_name[i] + ': ' + 'Detected'
                    text.textLine(line)
                    continue
                if key_name[i] == 'Staining' or 'Water pooling' or 'Other damage':
                    line = key_name[i] + ': ' + 'Major'
                    text.textLine(line)
                    continue
                if key_name[i] == 'Skylight' or 'HVAC/Cooling tower':
                    line = key_name[i] + ': ' + 'Many'
                    text.textLine(line)
                    continue

        pdf.drawText(text)
    ################################################## 19 points
        # if row[key_name_index]!=0:
        #     line= f'Total points: {int(row[key_name_index])}'
        #     text.textLine(line)
        # else:
        #     line = 'Total points: 0'
        #     text.textLine(line)
        # for i in range(key_name_index+1, len(key_name)):
        #     if row[i]>0:
        #         line = key_name[i] + ': ' + str(int(row[i]))
        #         text.textLine(line)
        # pdf.drawText(text)
    ##########################################################

        # images
        textLines = ['Image',
                    'Damage',
                    'Equipment'
        ]
        image_name = address + '.jpg'
        image_full_name = os.path.join(demoFolder,image_name)

        image = cv2.imread(image_full_name)
        height = image.shape[0]
        width = image.shape[1]
        ratio = width/height

        if height < width: # height < width horizontal
            tmp_width = 350
            tmp_height = 180
            new_width = ratio * tmp_height
            if new_width > tmp_width:
                draw_width = tmp_width
                draw_height = draw_width/ratio
            else:
                draw_width = new_width
                draw_height = tmp_height
            text = pdf.beginText(100, 500)
            text.setFont("Courier", 12)
            text.setFillColor(colors.blue)
            text.textLine(textLines[-3])
            pdf.drawText(text)
            pdf.drawInlineImage(image_full_name, 200, 440, width = draw_width, height = draw_height)
            pdf.linkURL(url, (200, 440, 200+draw_width, 440+draw_height), relative=1)
    # https://apps.nearmap.com/maps/#/@30.6622190,-88.1816001,20.00z,0d/V/20220110

            text = pdf.beginText(100, 300)
            text.setFont("Courier", 12)
            text.setFillColor(colors.blue)
            text.textLine(textLines[-2])
            pdf.drawText(text)
            pdf.drawInlineImage(Damage_image_name, 200, 240, width = draw_width, height = draw_height)
            pdf.linkURL(url, (200, 240, 200+draw_width, 240+draw_height), relative=1)


            text = pdf.beginText(100, 100)
            text.setFont("Courier", 12)
            text.setFillColor(colors.blue)
            text.textLine(textLines[-1])
            pdf.drawText(text)
            pdf.drawInlineImage(Equipment_image_name, 200, 40, width = draw_width, height = draw_height)
            pdf.linkURL(url, (200, 40, 200+draw_width, 40+draw_height), relative=1)

        else: # vertical
            tmp_width = 250
            tmp_height = 300
            new_width = ratio * tmp_height
            if new_width > tmp_width:
                draw_width = tmp_width
                draw_height = draw_width/ratio
            else:
                draw_width = new_width
                draw_height = tmp_height
            text = pdf.beginText(310, 370)
            text.setFont("Courier", 12)
            text.setFillColor(colors.blue)
            text.textLine(textLines[-3])
            pdf.drawText(text)
            pdf.drawInlineImage(image_full_name, 310, 380, width = draw_width, height = draw_height)
            pdf.linkURL(url, (310, 380, 310+draw_width, 380+draw_height), relative=1)

            text = pdf.beginText(40, 30)
            text.setFont("Courier", 12)
            text.setFillColor(colors.blue)
            text.textLine(textLines[-2])
            pdf.drawText(text)
            pdf.drawInlineImage(Damage_image_name, 40, 40, width = draw_width, height = draw_height)
            pdf.linkURL(url, (40, 40, 40+draw_width, 40+draw_height), relative=1)


            text = pdf.beginText(310, 30)
            text.setFont("Courier", 12)
            text.setFillColor(colors.blue)
            text.textLine(textLines[-1])
            pdf.drawText(text)
            pdf.drawInlineImage(Equipment_image_name, 310, 40, width = draw_width, height = draw_height)
            pdf.linkURL(url, (310, 40, 310+draw_width, 40+draw_height), relative=1)


        # saving the pdf
        pdf.save()

def zipdir(folder, path, ziph):
    file_path = os.path.join(folder, path)
    # ziph is zipfile handle
    if not os.path.isdir(file_path):
        if path[-3:]=='jpg': # move the orignial image to image folder
            ziph.write(file_path,path.replace('.jpg','/Image.jpg'))
        else:
            ziph.write(file_path, path)
    else:
        for root, dirs, files in os.walk(file_path):           
            for file in files:
                if file[0:4] != 'Roof': # don't include the roof boundary model
                    ziph.write(os.path.join(root, file), os.path.join(path, file))
                    
def zip_download_file(file_name):
    ############################## ZIP
    # excel_file_path = "roof_score_new_20220803.csv"
    # excel_file_path = "roof_score_new_20220726.csv"
    # image_folder_name = os.path.splitext(os.path.basename(excel_file_path))[0]

    # import glob
    # jpgFilenamesList = glob.glob(image_folder_name + '/*.jpg')
    # pdfFilenamesList = glob.glob(image_folder_name + '/*.pdf')
        # for f in jpgFilenamesList:   
        #     myzip.write(f)
        # for f in pdfFilenamesList:   
        #     myzip.write(f)
    image_folder_name = 'roof_score_new_20230101'

    file_list = [f for f in os.listdir(image_folder_name) if f[0].isnumeric()]
    print('Zipping')
    with zipfile.ZipFile(os.path.join(image_folder_name, file_name), 'w') as myzip:
        myzip.write(os.path.join(image_folder_name, 'all_address.csv'), 'all_address_' + image_folder_name[-8:] + '.csv')
        for f in file_list:    
            zipdir(image_folder_name, f, myzip)
    print('Zipping Done')

def send_email(receiver_email, address):
    host_server = 'smtp.qq.com'
    sender_qq = '1303242060'
    pwd = 'fhzevydopuvagfae'
    # pwd = input('Please input the password (not the login pwd)')

    sender_qq_mail = '1303242060@qq.com'
    # receiver = 'ruixu_liu@cinfin.com'

    smtp = SMTP_SSL(host_server)
    smtp.set_debuglevel(0)   # 1 for debug
    smtp.ehlo(host_server)
    smtp.login(sender_qq, pwd)


    msg = MIMEMultipart() # .txt .jpg .xls
    mail_title = 'Roof condition file'
    mail_content = address
    msg.attach(MIMEText(mail_content,'plain','utf-8'))

    # Add attachment
    att1=MIMEText(open('roof_score_new_20230101/roof.zip','rb').read(),'base64','utf-8')  #open file
    att1['Content-Type']='application/octet-stream' 
    att1['Content-Disposition']='attachment;filename=roof.zip'  

    # att2=MIMEText(open('aaa.jpg','rb').read(),'base64','utf-8')
    # att2['Content-Type']='application/octet-stream'
    # att2['Content-Disposition']='attachment;filename=aaa.jpg' 
    msg.attach(att1)
    # msg.attach(att2)

    print('Preparing done.')
    msg["Subject"] = Header(mail_title, 'utf-8')
    msg["From"] = sender_qq_mail
    msg["To"] = receiver_email

    smtp.sendmail(sender_qq_mail, receiver_email, msg.as_string())

    smtp.quit()
    print('Finish send email.')


import argparse
import os
import io
import sys
import json 
import numpy as np
import pandas as pd
import math
import cv2
import platform
import requests
import logging

import time
import timeit
from datetime import datetime

# from pathlib import Path
# plot
from PIL import Image
from matplotlib import pyplot as plt

### coco mask to polygon ###
import base64
from pycocotools.coco import COCO
from skimage import measure
from io import BytesIO

### parallel download images ###
import uuid
from joblib import Parallel, delayed
from concurrent import futures

#### shapely ###
import shapely
import shapely.wkt
from shapely.ops import unary_union
from shapely.geometry import Polygon, MultiPoint
from shapely.geometry import Polygon

### torch ###
import torch
import torch.backends.cudnn as cudnn

### detectron2 ###
from detectron2.data import DatasetCatalog
from detectron2.data.datasets.coco import register_coco_instances 
from detectron2.data.detection_utils import read_image
from detectron2.data import detection_utils as utils
from detectron2.utils.logger import setup_logger
from detectron2.utils.visualizer_new import Visualizer
from detectron2.utils.visualizer_new import ColorMode
from detectron2.engine.defaults import DefaultPredictor
from detectron2.data import MetadataCatalog

### yolo ###
from models.common import DetectMultiBackend
from utils.dataloaders import IMG_FORMATS, VID_FORMATS, LoadImages, LoadStreams
from utils.general import (LOGGER, Profile, check_file, check_img_size, check_imshow, check_requirements, colorstr, cv2,
                           increment_path, non_max_suppression, print_args, scale_coords, strip_optimizer, xyxy2xywh)
from utils.plots import Annotator, save_one_box
from utils.plots import colors as yolo_colors
from utils.segment.general import process_mask, scale_masks
from utils.segment.plots import plot_masks
from utils.torch_utils import select_device, smart_inference_mode
from utils.augmentations import letterbox

import yolo_roof.export

################################################### yolo model
def load_yolo_model(
        weights='best.pt',  # model.pt path(s)
        data='None',  # dataset.yaml path
        device='cpu',  # cuda device, i.e. 0 or 0,1,2,3 or cpu
        half=False,  # use FP16 half-precision inference
        dnn=False,  # use OpenCV DNN for ONNX inference
):
    # Load model
    device = select_device(device)
    print(device)
    model = DetectMultiBackend(weights, device=device, dnn=dnn, data=data, fp16=half)
    return model


def yolo_predict_one_image(
    yolo_predict_model,
    im0,
    imgsz=(1984, 1984),  # inference size (height, width)
    conf_thres=0.3,  # confidence threshold
    iou_thres=0.45,  # NMS IOU threshold
    max_det=100,  # maximum detections per image
    device='cpu',  # cuda:0 cuda device, i.e. 0 or 0,1,2,3 or cpu
    view_img=False,  # show results
    save_txt=False,  # save results to *.txt
    save_conf=False,  # save confidences in --save-txt labels
    save_crop=False,  # save cropped prediction boxes
    nosave=False,  # do not save images/videos
    classes=None,  # filter by class: --class 0, or --class 0 2 3
    agnostic_nms=False,  # class-agnostic NMS
    augment=True,  # augmented inference
    visualize=False,  # visualize features
    update=False,  # update all models
    project='runs/predict-seg',  # save results to project/name
    name='exp',  # save results to project/name
    exist_ok=True,  # existing project/name ok, do not increment
    line_thickness=3,  # bounding box thickness (pixels)
    hide_labels=False,  # hide labels
    hide_conf=False,  # hide confidences
    half=False,  # use FP16 half-precision inference
    dnn=False,  # use OpenCV DNN for ONNX inference
):
    stride, names, pt = yolo_predict_model.stride, yolo_predict_model.names, yolo_predict_model.pt
    imgsz = check_img_size(imgsz, s=stride)  # check image size

    bs = 1
    im = letterbox(im0, imgsz, stride=stride, auto=pt)[0]  # padded resize
    im = im.transpose((2, 0, 1))[::-1]  # HWC to CHW, BGR to RGB
    im = np.ascontiguousarray(im)
    # Run inference
    yolo_predict_model.warmup(imgsz=(1 if pt else bs, 3, *imgsz))  # warmup

    im = torch.from_numpy(im).to(device)
    im = im.half() if yolo_predict_model.fp16 else im.float()  # uint8 to fp16/32
    im /= 255  # 0 - 255 to 0.0 - 1.0
    if len(im.shape) == 3:
        im = im[None]  # expand for batch dim

    # Inference
    pred, out = yolo_predict_model(im, augment=False, visualize=False)
    proto = out[1]

    # NMS
    pred = non_max_suppression(pred, conf_thres, iou_thres, classes, agnostic_nms, max_det=max_det, nm=32)

    # Second-stage classifier (optional)
    # pred = utils.general.apply_classifier(pred, classifier_model, im, im0s)

    # Process predictions
    # for i, det in enumerate(pred):  # per image
    det = pred[0]
    im0 = im0.copy()
    annotator = Annotator(im0, line_width=line_thickness, example=str(names))
    scaled_masks_poly= []
    scores = []
    classes= []

    if len(det):
        masks = process_mask(proto[0], det[:, 6:], det[:, :4], im.shape[2:], upsample=True)  # HWC
        masks_poly = torch.as_tensor(masks, dtype=torch.uint8)
        masks_poly = masks_poly.permute(1, 2, 0).contiguous()
        masks_poly = masks_poly.cpu().numpy()
        scaled_masks_poly = scale_masks(im.shape[2:], masks_poly, im0.shape)

        total_masks = scaled_masks_poly.shape[2]
        # for idx in range(total_masks):
        #     mask = scaled_masks_poly[:,:, idx]
        #     segmentation = binary_mask_to_polygon(mask, tolerance=3)

        # Rescale boxes from img_size to im0 size
        # det[:, :4] = scale_coords(im.shape[2:], det[:, :4], im0.shape).round()

        scores = det[:, 4].cpu().numpy()
        classes = det[:, 5].cpu().numpy()

        # Mask plotting ----------------------------------------------------------------------------------------
        mcolors = [yolo_colors(int(cls), True) for cls in classes]
        im_masks = plot_masks(im[0], masks, mcolors)  # image with masks shape(imh,imw,3)
        annotator.im = scale_masks(im.shape[2:], im_masks, im0.shape)  # scale to original h, w
        # Mask plotting ----------------------------------------------------------------------------------------


        # # Write results
        # for *xyxy, conf, cls in reversed(det[:, :6]):
        #     c = int(cls)  # integer class
        #     label = None if hide_labels else (names[c] if hide_conf else f'{names[c]} {conf:.2f}')
            # annotator.box_label(xyxy, label, color=colors(c, True))
            # # Stream results

        im0 = annotator.result()
        # cv2.imwrite('aaa.jpg',im0)
    return scaled_masks_poly, scores, classes, im0


def mask_prediction(yolo_model, input_files, trainingDataset , output_folder, cut_image = False,tile_len = 800,tile_overlap = 0):
    '''
    Saved models do NOT contain prediction category descriptions or names, but only a category number. 
    To make sure the labels are correct, we need to use the metadata from the training dataset for this model.
    '''
    logger = setup_logger()    
    tile_image_length = tile_len
    overlap = tile_overlap
    boundary_shift = 0
    
    metadata = MetadataCatalog.get(trainingDataset)
    logging.info('training classes:')
    logging.info(metadata.get("thing_classes", None))

    categories = [{'id': idx, 'name': value,"supercategory": trainingDataset} for idx, value in enumerate(metadata.thing_classes)]
    logging.info(categories)
    
    model_folder = os.path.join(output_folder,'prediction_' + trainingDataset)
    if not(os.path.isdir(model_folder)):
        os.makedirs(model_folder)
        
    new_json_file = trainingDataset + '_all_in_one.json'
    new_annotations = []
    new_images = []
    ann_id = 1
    new_image_id = 1
    
    if input_files:
        for image_path in input_files:
            image_folder = os.path.splitext(image_path)[0]
            if not(os.path.isdir(image_folder)):
                os.makedirs(image_folder)

            img = read_image(image_path, format="BGR")
            
            if os.path.isdir(output_folder):
                out_filename = os.path.join(image_folder,trainingDataset +'.jpg')
                out_filename2 = os.path.join(model_folder, os.path.basename(image_path))
                out_filename2_extension = os.path.splitext(out_filename2)[1]
            
            image_name = os.path.basename(image_path)
            
            all_masks = []
            all_scores = []
            all_labels = []

            start_time = time.time()
            
            if cut_image:
                # slice cutting on the raw image
                raw_image = img.copy()
                output_image = img.copy()
                
                # compute total tiles
                #for corner case 1500 tilelength 800 overlap 100 should be 2 tiles while 1501 should be 3
                if raw_image.shape[0]<=tile_image_length:
                    total_i = 0
                else:
                    total_i = int((raw_image.shape[0]-1 - tile_image_length)  / (tile_image_length - overlap))+1
                if raw_image.shape[1]<=tile_image_length:
                    total_j = 0
                else:
                    total_j = int((raw_image.shape[1]-1 - tile_image_length) / (tile_image_length - overlap))+1


                total_instances = 0
                # loop for each tile i for width and j for longth
                for j in range(total_j + 1):
                    for i in range(total_i + 1):
                        idx = j * (total_i + 1) + i
                        # start of shape[0], end of shape[0], start of shape[1], end of shape[1]
                        if i == total_i:
                            shape0_start = max(boundary_shift, raw_image.shape[0] - tile_image_length)
                            shape0_end = raw_image.shape[0] - boundary_shift
                        else:
                            shape0_start = boundary_shift + (tile_image_length - overlap) * i
                            shape0_end = shape0_start + tile_image_length

                        if j == total_j:
                            shape1_start = max(boundary_shift, raw_image.shape[1] - tile_image_length)
                            shape1_end = raw_image.shape[1] - boundary_shift
                        else:
                            shape1_start = boundary_shift + (tile_image_length - overlap) * j
                            shape1_end = shape1_start + tile_image_length

                        # shape0_start, shape0_end, shape1_start, shape1_end = tile_image_coordinate[idx]
                        tile_image = (raw_image[shape0_start: shape0_end, shape1_start:shape1_end, :]).copy()

                        # visualized_output, polygonTuple = predVisSingleImage(predictor, tile_image, metadata)
                        scaled_masks_poly, scores, classes, im0 = yolo_predict_one_image(yolo_model,tile_image,imgsz=(tile_image.shape[0],tile_image.shape[1]))
                        
                        output_image[shape0_start: shape0_end, shape1_start:shape1_end, :] = im0
                        
                        total_polygons = len(scores)
                        if total_polygons:
                            for idx in range(total_polygons):
                                mask = scaled_masks_poly[:,:,idx]
                                segmentation = binary_mask_to_polygon(mask.astype(int), tolerance=0)
                                for item in segmentation:
                                    if len(item) > 6:
                                        list1 = []
                                        for point_i in range(0, len(item), 2):
                                            list1.append([shape1_start + item[point_i], shape0_start + item[point_i + 1]])

                                        current_polygon = np.array(list1)

                                        all_masks.append(Polygon(list1))
                                        all_scores.append(scores[idx])
                                        all_labels.append(int(classes[idx]))                                            
                                                                   

                        total_instances += total_polygons
                logger.info(
                    "{}: detected {} instances in {:.2f}s".format(
                        image_path, total_instances, time.time() - start_time
                    )
                )

                for each_class_id in set(all_labels):
                    each_class_masks = [each_mask for each_mask, each_label in zip(all_masks, all_labels) if each_label == each_class_id and each_mask.is_valid]
                    union_masks = unary_union(each_class_masks)
                    label_names = metadata.get("thing_classes", None)
                    # print(type(union_masks))
                    if union_masks.geom_type == 'MultiPolygon':
                        for current_mask in union_masks:
                            current_polygon_list = current_mask.exterior.coords
                            current_polygon_list_xy = current_mask.exterior.coords.xy

                            coco_segmentation_format = np.array(current_mask.exterior.coords).reshape(1, len(current_mask.exterior.coords) * 2).tolist()
                            # append annotation to coco jason file list
                            new_annotations.append({"id": ann_id,
                                                    "image_id": new_image_id,
                                                    "category_id": each_class_id,
                                                    "iscrowd": 0,
                                                    "segmentation": coco_segmentation_format
                                                    })
                            # get new annotation id
                            ann_id += 1
                            
                    else:
                        current_polygon_list = union_masks.exterior.coords
                        current_polygon_list_xy = union_masks.exterior.coords.xy

                        coco_segmentation_format = np.array(current_polygon_list).reshape(1, len(union_masks.exterior.coords) * 2).tolist()
                        # append annotation to coco jason file list
                        new_annotations.append({"id": ann_id,
                                                "image_id": new_image_id,
                                                "category_id": each_class_id,
                                                "iscrowd": 0,
                                                "segmentation": coco_segmentation_format
                                                })
                        # get new annotation id
                        ann_id += 1
                
                # append image information to coco jason file list
                new_images.append({"file_name": image_name,
                                   "id": new_image_id,
                                   "height": raw_image.shape[0],
                                   "width": raw_image.shape[1]})
                # get new image id
                new_image_id += 1               
                
                # save output image
                # cv2.imwrite(out_filename,output_image) #replace by the merge_polygones
                cv2.imwrite(out_filename2,output_image) 
            
            else:
                # visualized_output, polygonTuple = predVisSingleImage(predictor, img, metadata)
                scaled_masks_poly, scores, classes, im0 = yolo_predict_one_image(yolo_model, img,imgsz=(tile_len, tile_len))
                
                total_polygons = len(scores)
                if total_polygons:
                    for idx in range(total_polygons):
                        mask = scaled_masks_poly[:,:,idx]
                        segmentation = binary_mask_to_polygon(mask.astype(int), tolerance=0)
                        for item in segmentation:
                            if len(item) > 6:
                                list1 = []
                                for point_i in range(0, len(item), 2):
                                    list1.append([item[point_i], item[point_i + 1]])

                                current_polygon = np.array(list1)

                                all_masks.append(Polygon(list1))
                                all_scores.append(scores[idx])
                                all_labels.append(int(classes[idx]))
                    
                for each_class_id in set(all_labels):
                    each_class_masks = [each_mask for each_mask, each_label in zip(all_masks, all_labels) if each_label == each_class_id and each_mask.is_valid]
                    union_masks = unary_union(each_class_masks)
                    label_names = metadata.get("thing_classes", None)
                    # print(type(union_masks))
                    if union_masks.geom_type == 'MultiPolygon':
                        for current_mask in union_masks:
                            if current_mask.area < 100:
                                continue
                            current_polygon_list = current_mask.exterior.coords
                            current_polygon_list_xy = current_mask.exterior.coords.xy

                            coco_segmentation_format = np.array(current_mask.exterior.coords).reshape(1, len(current_mask.exterior.coords) * 2).tolist()
                            # append annotation to coco jason file list
                            new_annotations.append({"id": ann_id,
                                                    "image_id": new_image_id,
                                                    "category_id": each_class_id,
                                                    "iscrowd": 0,
                                                    "segmentation": coco_segmentation_format
                                                    })
                            # get new annotation id
                            ann_id += 1
                    else:
                        if union_masks.area < 100:
                            continue
                        current_polygon_list = union_masks.exterior.coords
                        current_polygon_list_xy = union_masks.exterior.coords.xy

                        coco_segmentation_format = np.array(current_polygon_list).reshape(1, len(union_masks.exterior.coords) * 2).tolist()
                        # append annotation to coco jason file list
                        new_annotations.append({"id": ann_id,
                                                "image_id": new_image_id,
                                                "category_id": each_class_id,
                                                "iscrowd": 0,
                                                "segmentation": coco_segmentation_format
                                                })
                        # get new annotation id
                        ann_id += 1
                # save output image
                cv2.imwrite(out_filename,im0) 
                cv2.imwrite(out_filename2,im0) 
                # visualized_output.save(out_filename)
                # visualized_output.save(out_filename2) 

                # append image information to coco jason file list
                new_images.append({"file_name": image_name,
                                   "id": new_image_id,
                                   "height": img.shape[0],
                                   "width": img.shape[1]})
                # get new image id
                new_image_id += 1 

                # output_image = visualized_output.get_image()[:, :, ::-1]
  
            
        info = {"year": 2022,
                "version": "1.0",
                "description": "Cutting the large image into small pieces and selecting desired classes",
                "contributor": "Ruixu Liu, Delin Shen, Aaron Lee and Qiang Wang",
                }

        json_data = {"info": info,
                     "categories": categories,
                     "images": new_images,
                     "annotations": new_annotations
                     }

        with open(os.path.join(model_folder,new_json_file), "w") as jsonfile:
            json.dump(json_data, jsonfile, sort_keys=True, indent=4)
        
        vis_out = merge_polygones(output_folder,os.path.join(model_folder,new_json_file),trainingDataset)
        output_image = vis_out.get_image()[:, :, ::-1]
        return output_image


def close_contour(contour):
    if not np.array_equal(contour[0], contour[-1]):
        contour = np.vstack((contour, contour[0]))
    return contour

def binary_mask_to_polygon(binary_mask, tolerance=0):
    """Converts a binary mask to COCO polygon representation
    Args:
        binary_mask: a 2D binary numpy array where '1's represent the object
        tolerance: Maximum distance from original points of polygon to approximated
            polygonal chain. If tolerance is 0, the original coordinate array is returned.
    """
    polygons = []
    # pad mask to close contours of shapes which start and end at an edge
    padded_binary_mask = np.pad(binary_mask, pad_width=1, mode='constant', constant_values=0)
    contours = measure.find_contours(padded_binary_mask, 0.5)
    contours = np.subtract(contours, 1)
    for contour in contours:
        contour = close_contour(contour)
        contour = measure.approximate_polygon(contour, tolerance)
        if len(contour) < 3:
            continue
        contour = np.flip(contour, axis=1)
        segmentation = contour.ravel().tolist()
        # after padding and subtracting 1 we may get -0.5 points in our segmentation
        segmentation = [0 if i < 0 else i for i in segmentation]
        polygons.append(segmentation)
    return polygons

def img_tobyte(im_cv):
    im_rgb = cv2.cvtColor(im_cv, cv2.COLOR_BGR2RGB)
    img_pil = Image.fromarray(im_rgb)
    ENCODING = 'utf-8'
    img_byte = io.BytesIO()
    img_pil.save(img_byte, format='JPEG')
    binary_str2 = img_byte.getvalue()
    imageData = base64.b64encode(binary_str2)
    base64_string = imageData.decode(ENCODING)
    return base64_string

def merge_polygones(image_folder,json_file,trainingDataset):
    allColors = [(255,0,0), (0,255,0), (0,0,255), (255,255,0), (255,0,255), (0,255,255), (75,125,255), (125,75,255), (75,255,125), (125,255,75), (255,125,75), (255,75,125)]*8
    filename = str(uuid.uuid4())
    register_coco_instances(filename, {'thing_colors':allColors}, json_file, image_folder)
    metadata = MetadataCatalog.get(filename)
    dicts = DatasetCatalog.get(filename)
    model_folder = os.path.join(image_folder,'prediction_' + trainingDataset)
    logging.info(f"Merging {trainingDataset} polygons")
    for dic in dicts:
        img = utils.read_image(dic["file_name"], "RGB")
        visualizer = Visualizer(img, metadata=metadata, instance_mode=ColorMode.SEGMENTATION)
        vis = visualizer.draw_dataset_dict(dic)
        out_filename = os.path.join(os.path.splitext(os.path.basename(dic["file_name"]))[0],trainingDataset +'_merged.jpg')
        
        out_filename2_extention = os.path.splitext(os.path.basename(dic["file_name"]))[1]
        out_filename2 = os.path.basename(dic["file_name"]).replace(out_filename2_extention,'_merged'+out_filename2_extention)
        vis.save(os.path.join(image_folder,out_filename))
        # vis.save(os.path.join(model_folder,out_filename2)) # model folder is not necessary
    logging.info("Merging Finished.")
    return vis

def translate(classes, scores, class_names):
    labels = None
    if classes is not None:
        if class_names is not None and len(class_names) > 0:
            labels = [class_names[i] for i in classes]
        else:
            labels = [str(i) for i in classes]
    if scores is not None:
        if labels is None:
            labels = ["{:.0f}%".format(s * 100) for s in scores]
        else:
            labels = ["{} {:.0f}%".format(l, s * 100) for l, s in zip(labels, scores)]
    return labels

def inputimage_list(input_folder):
    files = os.listdir(input_folder)
    input_files = [os.path.join(input_folder, x) for x in files if os.path.splitext(x)[1] in ['.jpg','.png']]
    return input_files


def roof_predict(yolo_roof_model, yolo_damage_model, yolo_equipment_model, file_root_path):
    demoFolder = 'roof_score_new_20230101'
    ############################################ input Image
    input_files = inputimage_list(input_folder = demoFolder)
    logging.info(input_files)
    ########################################## dataset config
    imgsetUrl =  os.path.join(file_root_path,'test_360.jpg')

    cocoJsonPairs = {
        'Roof_boundary' :  os.path.join(file_root_path,'roof_boundary.json'),
        'Roof_damage' :  os.path.join(file_root_path,'roof_damage.json'),
        'Roof_equipment' :  os.path.join(file_root_path,'roof_equipment.json'),
    }

    for datasetName, jsonURL in cocoJsonPairs.items():     
        if datasetName not in DatasetCatalog:
            # there is a way to load the labels from the detectron2 datasets. should replace this piece.
            with open(jsonURL) as f:
                cocoFile = json.load(f)
            catMap = {str(cat['id']): cat['name'] for cat in cocoFile['categories']}  # coco file category ID starts from 1
            logging.info(catMap)
            catIDs = [int(key) for key in catMap.keys()] # when detectron2 load a coco dataset, the categories are sorted by IDs, and mapped to [1, lengthOfCatetories] as a subset of [1,80).

            # To make sure the labels are in the same order as the IDs when passed to register the dataset
            catIDs.sort() 
            labels = [catMap[str(catID)] for catID in catIDs]
            allColors = [(255,0,0), (0,255,0), (0,0,255), (255,255,0), (255,0,255), (0,255,255), (75,125,255), (125,75,255), (75,255,125), (125,255,75), (255,125,75), (255,75,125)]*8 
            # 12 colors; repeat 8 times in case there are more labels than 12; need a better way to assign the colors
            colors = allColors[:len(labels)]
            logging.info('dataset: %s has %i labels and %i colors ' % (datasetName, len(labels), len(colors)))
            register_coco_instances(datasetName, {'thing_classes':labels, 'thing_colors':colors}, jsonURL, imgsetUrl) 
            # the 2nd parameter is metadata, in this case we pass over thing_classes for Visualizer to use    
        else: 
            logging.info('%s already registered!' % datasetName)

    ############################################# Models
    last_roof_boundary_result = mask_prediction(yolo_roof_model, input_files, trainingDataset='Roof_boundary', output_folder=demoFolder,  cut_image = False, tile_len = 1984, tile_overlap = 500)
    last_damage_result = mask_prediction(yolo_damage_model, input_files, trainingDataset='Roof_damage', output_folder=demoFolder,  cut_image = True, tile_len = 800, tile_overlap = 100)
    last_equipment_result = mask_prediction(yolo_equipment_model, input_files, trainingDataset='Roof_equipment', output_folder=demoFolder,  cut_image = True, tile_len = 800, tile_overlap = 100)
